#   request.py: request system code for IDJC
#   Copyright (C) 2007 Stephen Fairchild (s-fairchild@users.sourceforge.net)
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program in the file entitled COPYING.
#   If not, see <http://www.gnu.org/licenses/>.

import pygtk
pygtk.require('2.0')
import gtk, os, socket, time, sys, select, zlib
from langheader import *

"""
Protocol guide for interface writers:

   1. Create and connect a socket to the receive port being used by IDJC (6240 default)
   2. Send the command
   3. Shut down the write end of the socket
   4. Read the reply
   5. Close the socket

The command format is as follows:
   
   idjccomms:A:C:x
   
   A is the protocol number which is currently always "01"
   
   C is the command which can be one of
   
      getdb      - obtain the database of requestable files
      getqueue   - obtain the list of queued requests
      request    - make a request
      request_at - request by artist - title
      accepting  - enquire if IDJC is currently accepting requests
      		  also serves as a general purpose ping function
      
   x pertains to the additional parameters specific to certain commands:
      
      The 'reqeust' command has x defined as:
         
         L:M:dS:sdT:tdU:u
         
         L is an 8 digit ASCII encoded hexadecimal CRC32 value of everything
           after the second colon
         M is the length of everything after the second colon in bytes
           it is 6 digits long and leading zero padded with ASCII encoding
         Both L and M can be omitted by substituting 'X' characters.
         
         d is literally the letter 'd' encoded in ASCII ***
         S is the length of the pathname in bytes
         s is the pathname data e.g. /home/jim/music/musicfile.mp3
         T is the length of the name of the person who made the request
         t is the name of the person who made the request
         U is the length of the dedication
         u is the text of the dedication
         
         T and U can both be ascii '0' to omit the t and u fields.
         
         The request_at command uses the same format but instead of a pathname
         you would send something like "Metallica - One" for 's'.
         
      The getdb command has x defined as:
         
         L:M:dS:s
         
         L and M are as defined the same as for 'request'
         
         S is the filter string length
         s is the filter string - not NULL terminated
         
         The filter string filters the database on 'Artist - Title' so that only
         a matching subset of the database is returned. It is case sensitive.
         By setting S to zero and omitting s you can obtain the entire database
         which could conceivably in a rare case exceed 1MB compressed.
         Example usage: idjccomms:01:getdb:XXXXXXXX:000005:d2:Me
         The above example will retrieve songs by Metallica but not by Coldplay.

   The trailing colon for all commands not requiring an x parameter is required.
 
The following replies are currently defined:
   
   00:(response of 'okay' or reply data follows)
   01:not accepting requests
   02:bad crc
   03:bad length
   04:unrecognised protocol
   05:request missing from request playlist
   06:command not recognised
   07:malformed data
   08:user blocked
   
   If the 'idjccomms:' prefix is not received then no reply will be sent.

The reply in the case of the getdb command is formatted as follows:
   
   00:L:M:N:P:dS:sdT:tdU:udV:v
   
   00 indicates success
   L is the CRC32 of everything from second colon onwards (8 hex digits in ASCII)
   M is the length in bytes of the data following the fifth colon (8 ASCII digits)
   N is the uncompressed length (see above) (8 ASCII)
   P is the number of rows of data (6 ASCII)
   Where there is no compression N will be filled with ASCII 'X' otherwise
   zlib compression is used following the fourth colon.
   
   The following part of the reply will repeated as many times as the value P
   for each track in the playlist.

   d is literally the letter 'd' encoded in ASCII
   S is the length of the pathname of the song
   s is the text of the pathname of the song
   T is the length of the string representing song length
   t is the song length string, ASCII encoded and in whole seconds
   U is the length of the song title text
   u is the song title which may also contain the artist in "artist - title" format
   V is the length of the artist text
   v is the artist text which may sometimes be blank
   
   Both u and v are UTF-8 encoded.
   The items in the list will be sorted alphabetically by Artist - Title.

The reply in the case of the getqueue command is formatted as:
   
   00:L:M:N:P:dS:sdT:tdU:udV:vdW:w
   
   00 indicates success
   L is the CRC32 of everything following the third colon (8 hex digits in ASCII)
   M is the length of everything from P onwards (8 ASCII)
   N is the uncompressed length of everything from P onwards (8 ASCII)
   P is the number of requests currently in the queue (6 ASCII)
   Where there is no compression, N will be filled with ASCII 'X', otherwise
   zlib compression is used from P onwards.

   The following part of the reply will be repeated as many times as the value for P
   for each pending request.

   d is literally the letter 'd' encoded in ASCII
   S is the title length
   s is the title text    - may also contain artist information
   T is the artist length - can be zero
   t is the artist text
   U is the requestor's name length - possibly '0'
   u is the requestor's name
   V is the dedication length	    - possibly '0'
   v is the dedication text
   W is the length of the duration text
   w is the duration text in whole seconds as ASCII
   
   This feature is needed since the DJ is free to delete requests or
   alter the play order.
   
The replies for the 'request' command:
   
   00:okay
   
   or 
   
   01:not accepting requests
   08:user blocked
   
   A code of 01 indicates the DJ has opted to block all requests
   A code of 08 indicates the DJ had barred a particular user
   
The reply for the 'accepting' command will be:
   
   00:okay
   
   or 
   
   01:not accepting requests


*** Note that 'd' precedes a probably multi character ASCII encoded number, terminated
    by the first ':' character encountered. This is to be treated as a base 10 number
    representing a string length, with the string to follow. The strings are not null 
    terminated.

"""

def hexcrc32(data):	# Python returns CRC as an int - bleh
    crc = zlib.crc32(data)
    res=''
    for i in range(4):
        t=crc & 0xFF
        crc >>= 8
        res='%02X%s' % (t, res)
    return res

def extractor(data):
   i = 1
   while 1:
      if data[i - 1] != "d":
         print "request.extractor: WARNING, read past end boundary"
         yield None
         continue
      colon_index = data.index(":", i)
      text_length = int(data[i : colon_index])
      text = data[colon_index + 1 : colon_index + 1 + text_length]
      yield text
      i = colon_index + text_length + 2 

class RequestSystem(gtk.HBox):
   requests_database = None
   requests_db_index = None
   requests_queue = None

   class Handler:
      def handle_command_01(self, command):
         def LM_chk(command):
            if len (command) < 16 or command[8] != ":" or command[15] != ":":
               return "07:malformed data"
            asccrc = command[:8]
            asclen = command[9:15]
            if asclen != "XXXXXX":
               try:
                  intlen = int(asclen)
               except ValueError:
                  return "03:bad length"
               if intlen + 16 != len(command):
                  return "03:bad length"
            if asccrc != "XXXXXXXX" and asccrc != hexcrc32(command[16:]):
               return "02:bad crc"
            return None

         def get_requests_database(rs, player, command):
            print "get_requests_database was called"
            if player.playlist_changed:
               print "making a new database"
               # make a new database collecting relevant info, removing duplicates,
               # sorting, then creating an index for easy searching
               database = []
               for row in player.liststore:
                  # pathname, play duration, title, artist, artist - title
                  line = (row[1], str(row[2]), row[5], row[6], row[3])
                  database.append(line)
               print "unsorted database", database
               database.sort(lambda x, y: cmp(x[1], y[1]))
               for index in range(len(database) -1):
                  if database[index][0] == database[index + 1][0]:
                     database[index] = None
               rs.requests_database = []
               for each in database:
                  if each is not None:
                     rs.requests_database.append(each)
               rs.requests_database.sort(lambda x, y: cmp(x[4], y[4]))
               print "sorted database", rs.requests_database
               rs.requests_db_index = {}
               oldlongkey = ""
               oldshortkey = ""
               for index, each in enumerate(rs.requests_database):
                  shortkey = str(ord(each[4][0]))
                  longkey = "%s,%d" % (shortkey, ord(each[4][1]))
                  if shortkey != oldshortkey:
                     rs.requests_db_index[shortkey] = index
                  if longkey != oldlongkey:
                     rs.requests_db_index[longkey] = index
               print "database index", rs.requests_db_index
               player.playlist_changed = False
            print "extracting from", command
            prefix = extractor(command).next()
            if prefix is None:
               reply = "07:malformed data"
            else:
               replydata = []
               if prefix == "":
                  for row in rs.requests_database:
                     replydata.append("d%d:%sd%d:%sd%d:%sd%d:%s" % (len(row[0]), row[0],
                                                                    len(row[1]), row[1],
                                                                    len(row[2]), row[2],
                                                                    len(row[3]), row[3]))
               else:
                  key = ",".join(( str(ord(each)) for each in prefix[:2] ))
                  ix = requests_db_index[key]
                  try:
                     while 1:
                        if rs.requests_database[ix][3].startswith(prefix):
                           break
                        ix += 1
                  except IndexError:
                     pass
                  else:
                     for row in rs.requests_database[ix:]:
                        if row[3].startswith(prefix):
                           replydata.append("d%d:%sd%d:%sd%d:%sd%d:%s" % (len(row[0]), row[0],
                                                                  len(row[1]), row[1],
                                                                  len(row[2]), row[2],
                                                                  len(row[3]), row[3]))
                        else:
                           break
               replydatastr = "".join(replydata)
               if len(replydatastr) > 1300:
                  uncompressed_size = "%08d" % len(replydatastr)
                  replydatastr = zlib.compress(replydatastr)
               else:
                  uncompressed_size = "XXXXXXXX"
               data_size = "%08d" % len(replydatastr)
               rows = "%06d" % len(replydata)
               crcdata = "%s:%s:%s:%s" % (data_size, uncompressed_size, rows, replydatastr)
               crc32value = hexcrc32(crcdata)
               reply = "00:%s:%s" % (crc32value, crcdata)
            return reply

         def get_requests_queue(rs, player):
            print "get_requests_queue was called"
            return "00"

         def make_request(player, command, request_by):
            print "make_request was called"
            return "00"
         
         def accepting_requests(player):
            print "accepting_requests was called"
            return "01:requests are off"

         self.parent.console("received command: %s" % command)
         if command == "getqueue..:":
            self.write(get_requests_queue(self.parent, self.parent.activeplayer))
         elif command == "accepting.:":
            self.write(accepting_requests(self.parent.activeplayer))
         else:
            reply = LM_chk(command[11:])
            if reply is not None:
               if command.startswith("getdb.....:"):
                  self.write(get_requests_database(self.parent, self.parent.activeplayer, command[22:]))
               elif command.startswith("request...:"):
                  self.write(make_request(self.parent.activeplayer, command[24:], "path"))
               elif command.startswith("request_ap:"):
                  self.write(make_request(self.parent.activeplayer, command[27:], "a-t"))
	    else:
               self.write(reply)

      def remove_handler(self):
         for index, each in enumerate(self.parent.handlerlist):
            if each[1] == self:
               del self.parent.selectlist[self.parent.selectlist.index(each[0])]
               del self.parent.handlerlist[index]
               try:
                  each[0].close()
               except socket.error:
                  pass
      
      def write(self, string_to_send):
         try:
            self.conn.send(string_to_send)
            self.conn.shutdown(socket.SHUT_WR)
         except socket.error:
            self.parent.console("Reply to %s port %d failed" % self.addr)
         else:
            self.parent.console("Reply sent to %s port %d" % self.addr)
      
      def read(self):
         data = self.conn.recv(1024)
         if data:
            self.totaldata += data
            if len(self.totaldata) >= 10 and not self.totaldata.startswith("idjccomms:"):
               try:
                  self.parent.console("Pass phrase not received")
                  self.conn.shutdown(socket.SHUT_RDWR)
               except socket.error:
                  pass
               self.remove_handler()
         else:
            try:
               self.conn.shutdown(socket.SHUT_RD)
            except socket.error:
               self.remove_handler()
            else:
               if len(self.totaldata) >= 13 and self.totaldata[12] ==":":
                  protocol = self.totaldata[10:12]
                  if protocol == "01":
                     self.handle_command_01(self.totaldata[13:])
                  else:
                     self.parent.console("Unknown protocol: %s" % protocol)
               else:
                  self.parent.console("Received bad data")
            self.remove_handler()
      
      def __init__(self, parent, conn, addr):
         self.parent = parent
         self.conn = conn
         self.addr = addr
         self.totaldata = ""
   
   def console(self, text):
      self.text_buffer.insert(self.text_buffer.get_end_iter(), "".join((time.strftime("%H:%M"), ": ",text, "\n")))
      self.requests_text_view.scroll_to_iter(self.text_buffer.get_end_iter(), 0.0)

   def handle_listen_port(self):
      if self.listenpoll.poll(0):
         print "Poll succeeded on listen port"
         try:
            conn, addr = self.listensocket.accept()
         except:
            self.console("Connection attempt failed")
         else:
            self.console("New connection from %s on port %d" % addr)
            conn.setblocking(0)
   	    self.selectlist.append(conn)
            self.handlerlist.append((conn, self.Handler(self, conn, addr)))

   def handle_communications(self):
      try:
         readable, writable, error = select.select(self.selectlist, [], [], 0)
      except:
         for each in self.selectlist:
            try:
               select.select([each], [], [], 0)
            except:
               del self.selectlist[self.selectlist.index(each)]
               self.handle_communications()
               return
      for port in readable:
         for (conn, handler) in self.handlerlist:
            if port == conn:
               handler.read()

   def comms_pass(self):
      if self.activeplayer is not None:
         self.handle_listen_port()
         self.handle_communications()
   
   def get_active_player(self):
      if self.use_playlist1.get_active():
         return self.idjcgui.player_left
      else:
         return self.idjcgui.player_right
   
   def get_bind_setting(self):
      if self.bind_localhost.get_active():
         return "localhost"
      if self.bind_hostname.get_active():
         return socket.gethostname()
      return self.bind_entry.get_text().strip().encode("utf-8", "replace")
   
   def cb_activate(self, widget):
      if widget.get_active():
         try:
            self.listensocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.listensocket.bind((self.get_bind_setting(), int(self.request_port_adj.get_value())))
         except:
            try:
               del self.listensocket
            except:
               pass
            self.console("Problem connecting to the port")
            self.request_activate.set_active(False)
            return
         self.listensocket.setblocking(0)
         self.listensocket.listen(5)
         self.listenpoll = select.poll()
         self.listenpoll.register(self.listensocket)
         self.activeplayer = self.get_active_player()
         self.console("Started listening on port %d" % int(self.request_port_adj.get_value()))
      else:
         if self.activeplayer is not None:
            self.listenpoll.unregister(self.listensocket)
            del self.listenpoll
            self.listensocket.shutdown(socket.SHUT_RDWR)
            self.listensocket.close()
            del self.listensocket
            self.activeplayer = None
            self.console("Ceased listening on the request port")
            self.selectlist = []
            self.handlerlist = []
            self.console("Purged all open connections")

   def __init__(self, idjcgui):
      gtk.HBox.__init__(self)
      self.set_border_width(8)
      self.set_spacing(6)
      self.idjcgui = idjcgui
      self.activeplayer = None
      self.selectlist = []
      self.handlerlist = []
      
      lpanevbox = gtk.VBox()
      lpanevbox.set_spacing(3)
      self.pack_start(lpanevbox, False, False, 0)
      lpanevbox.show()
      
      frame = gtk.Frame(" " + port_text + " ")
      portvbox = gtk.VBox()
      portvbox.set_border_width(4)
      self.request_port_adj = gtk.Adjustment(6240.0, 1.0, 65535.0, 1.0, 1.0, 1.0)
      request_port_spin = gtk.SpinButton(self.request_port_adj, 0, 0)
      portvbox.add(request_port_spin)
      request_port_spin.show()
      frame.add(portvbox)
      portvbox.show()
      lpanevbox.add(frame)
      frame.show()
      
      sizegroup = gtk.SizeGroup(gtk.SIZE_GROUP_VERTICAL)
      frame = gtk.Frame(" " + bind_to_text + " ")
      bindbox = gtk.VBox()
      bindbox.set_border_width(4)
      frame.add(bindbox)
      bindbox.show()
      self.bind_localhost = gtk.RadioButton(None, localhost_text)
      sizegroup.add_widget(self.bind_localhost)
      bindbox.add(self.bind_localhost)
      self.bind_localhost.show()
      self.bind_hostname = gtk.RadioButton(self.bind_localhost, hostname_text)
      sizegroup.add_widget(self.bind_hostname)
      bindbox.add(self.bind_hostname)
      self.bind_hostname.show()
      other_hbox = gtk.HBox()
      sizegroup.add_widget(other_hbox)
      self.bind_other = gtk.RadioButton(self.bind_hostname)
      other_hbox.pack_start(self.bind_other, False, False, 0)
      self.bind_other.show()
      self.bind_entry = gtk.Entry()
      self.bind_entry.set_size_request(90, -1)
      other_hbox.pack_start(self.bind_entry, False, False, 0)
      self.bind_entry.show()
      bindbox.add(other_hbox)
      other_hbox.show()
      lpanevbox.add(frame)
      frame.show()
      
      frame = gtk.Frame(" " + use_text + " ")
      usebox = gtk.VBox()
      usebox.set_border_width(4)
      frame.add(usebox)
      usebox.show()
      self.use_playlist1 = gtk.RadioButton(None, playlist1_text)
      usebox.add(self.use_playlist1)
      self.use_playlist1.show()
      self.use_playlist2 = gtk.RadioButton(self.use_playlist1, playlist2_text)
      usebox.add(self.use_playlist2)
      self.use_playlist2.show()
      lpanevbox.add(frame)
      frame.show()
       
      self.request_activate = gtk.ToggleButton(request_activate_text)
      self.request_activate.connect("toggled", self.cb_activate)
      lpanevbox.add(self.request_activate)
      self.request_activate.show()
      
      requests_text_window = gtk.ScrolledWindow(None, None)
      requests_text_window.set_policy(gtk.POLICY_NEVER, gtk.POLICY_ALWAYS)
      requests_text_window.set_shadow_type(gtk.SHADOW_ETCHED_IN)
      self.pack_start(requests_text_window, True, True, 0)
      requests_text_window.show()
      self.requests_text_view = gtk.TextView()
      requests_text_window.add(self.requests_text_view)
      self.requests_text_view.show()
      self.requests_text_view.set_cursor_visible(False)
      self.requests_text_view.set_wrap_mode(gtk.WRAP_CHAR)
      self.requests_text_view.set_editable(False)
      self.text_buffer = self.requests_text_view.get_buffer()
